-- global variables
tilesize = 48
turn = 0

-- load jumper library for pathfinding
local Grid = require ("jumper.grid") -- The grid class
local Pathfinder = require ("jumper.pathfinder") -- The pathfinder class

-- init map
mapwidth = 15
mapheight = 10
map = {
	{2,2,2,2,2,2,2,2,2,2,2,2,2,2,2},
	{2,1,1,1,1,1,1,1,1,1,1,1,1,1,2},
	{2,1,1,2,2,2,2,2,1,1,1,1,1,1,2},
	{2,1,1,2,1,1,1,2,1,1,1,1,1,1,2},
	{2,1,1,1,1,1,1,2,1,1,1,1,1,1,2},
	{2,1,1,1,1,1,1,2,1,1,1,1,1,1,2},
	{2,1,1,2,1,1,1,2,1,1,1,1,1,1,2},
	{2,1,1,2,2,2,2,2,1,1,1,1,1,1,2},
	{2,1,1,1,1,1,1,1,1,1,1,1,1,1,2},
	{2,2,2,2,2,2,2,2,2,2,2,2,2,2,2}
	}

entities = {
	{type = "player", x = 2, y = 2, energy = 0, energygen = 100, movecost = 100}, -- player is always entity #1
	{type = "goblin", x = 14, y = 9, energy = 0, energygen = 100, movecost = 100},
	{type = "snake", x = 2, y = 9, energy = 0, energygen = 100, movecost = 80},
	{type = "troll", x = 14, y = 2, energy = 0, energygen = 100, movecost = 120}
	}

function love.load()
	love.keyboard.setKeyRepeat(false) -- disable key repeat
	
	-- load images for tiles and entities
	maptile = {}
	maptile[1] = love.graphics.newImage("1.png")
	maptile[2] = love.graphics.newImage("2.png")
	
	imgentity = {}
	imgentity.player = love.graphics.newImage("player.png")
	imgentity.goblin = love.graphics.newImage("goblin.png")
	imgentity.snake = love.graphics.newImage("snake.png")
	imgentity.troll = love.graphics.newImage("troll.png")
end

function love.update(dt)

end

function love.draw()
	-- draw map
	local a, b
	for a = 1, mapwidth do
		for b = 1, mapheight do
			love.graphics.draw(maptile[map[b][a]], (a - 1) * tilesize, (b - 1) * tilesize)
		end
	end
	
	-- draw entities
	local dx, dy
	local k, v
	for k, v in pairs(entities) do
		dx, dy = (v.x - 1) * tilesize, (v.y - 1) * tilesize
		love.graphics.draw(imgentity[v.type], dx, dy)
	end	
	
	-- print game information
	local str
	str = "Controls: left/down/up/right or hjkl to move, yubn for diagonal move, r to reset entities and . to pass turn.\n"
	str = str .. "Turns passed:" .. turn .. "\n"
	local k, v
	for k, v in pairs(entities) do
		str = str .. "Energy of " .. v.type .. ":" .. v.energy .. "\n"
	end	
	love.graphics.setColor(0, 0, 0)
	love.graphics.print(str, 0, 0)
	love.graphics.setColor(255, 255, 255)
end

function love.keypressed(key)
	local playertookaction = false
	-- h(←) j(↓) k(↑) l(→) y(↖) u(↗) b(↙) n(↘) p(perform pathfinding to mouse cursor)
	if key == "h" or key == "left" then -- move left
		playertookaction = move(player(), -1, 0)
	elseif key == "j" or key == "down" then -- move down
		playertookaction = move(player(), 0, 1)
	elseif key == "k" or key == "up" then -- move up
		playertookaction = move(player(), 0, -1)
	elseif key == "l" or key == "right" then -- move right
		playertookaction = move(player(), 1, 0)
	elseif key == "y" then -- move left-up
		playertookaction = move(player(), -1, -1)
	elseif key == "u" then -- move right-up
		playertookaction = move(player(), 1, -1)
	elseif key == "b" then -- move left-down
		playertookaction = move(player(), -1, 1)
	elseif key == "n" then -- move right-down
		playertookaction = move(player(), 1, 1)
	elseif key == "." then
		playertookaction = true
	elseif key == "r" then
		reset()
	elseif key == "p" then -- perform pathfinding to mouse pointer
		
	end
	
	if playertookaction == true then
		player().energy = player().energy - player().movecost
		while player().energy < 0 do
			player().energy = player().energy + player().energygen
			tick()
		end
	end
end

function tick()
	-- process turn and move another monsters
	turn = turn + 1
	
	-- make non-player creatures move
	local k, v, path
	local creature_moved
	for k, v in pairs(entities) do
		if v.type ~= "player" then
			-- perform pathfinding and track player
			v.energy = v.energy + v.energygen
			while(v.energy >= v.movecost) do
				path = pathfind(v, player())
				if path[1] ~= nil then
					move(v, path[1].x, path[1].y)
				end
				v.energy = v.energy - v.movecost
			end
		end
	end	
end

function player()
	return entities[1] -- player is always entity #1
end

function reset()
	-- reset turns and entities
	turn = 0
	entities = {
	{type = "player", x = 2, y = 2, energy = 0, energygen = 100, movecost = 100}, -- player is always entity #1
	{type = "goblin", x = 14, y = 9, energy = 0, energygen = 100, movecost = 100},
	{type = "snake", x = 2, y = 9, energy = 0, energygen = 100, movecost = 80},
	{type = "troll", x = 14, y = 2, energy = 0, energygen = 100, movecost = 120}
	}
end

function move(entity, mx, my)
	-- check if entity can move.
	-- if entity can move, move entity and return true.
	-- if entity can't move, return false.
	
	local newx, newy = entity.x + mx, entity.y + my
	if newx < 0 or newy < 0 or newx > mapwidth or newy > mapheight then
		return false -- can't move outside of the map boundary
	end
	if map[newy][newx] == 2 then
		return false -- can't move into non-walkable tile
	end
	
	local k, v
	for k, v in pairs(entities) do
		if newx == v.x and newy == v.y then
			-- tile is already occupied by another entity
			return false
		end
	end	
	
	-- entity can move to map[my][mx]
	entity.x = newx
	entity.y = newy
	
	return true
end



function pathfind(e1, e2)
	-- perform pathfinding between two entities and return path a array
	-- will return nil if pathfinder can't find path
	
	local gridmap = getwalkablemap(e1.x, e1.y, e2.x, e2.y)
	local grid = Grid(gridmap)
	local walkable = 1
	local myFinder = Pathfinder(grid, "ASTAR", walkable)
	local path, length = myFinder:getPath(e1.x, e1.y, e2.x, e2.y, true)
	local rpath = {} -- path in relative coordinates form
	if path then
		local a
		for a = #path, 2, -1 do
			rpath[a - 1] = {}
			rpath[a - 1].x = path[a].x - path[a - 1].x
			rpath[a - 1].y = path[a].y - path[a - 1].y
		end
	end
	return rpath
end

function getwalkablemap(x1, y1, x2, y2)
	-- get walkable grid map between x1:y1 and x2:y2
	-- get map
	local wmap = {}
	local a, b
	for a = 1, mapheight do
		wmap[a] = {}
		for b = 1, mapwidth do
			wmap[a][b] = map[a][b]
		end
	end
	
	-- mark entities except itself for non-walkable
	local k, v
	for k, v in pairs(entities) do
		if v ~= entity then -- do not include itself
			wmap[v.y][v.x] = "E"
		end
	end	
	
	-- assume start point and end point are always walkable
	wmap[y1][x1] = 1
	wmap[y2][x2] = 1
	
	return wmap
end